-- Filename:     moore_example.vhd
-- Filetype:     VHDL code (behavior)
-- Date:         25 april 2011
-- Update:       15 april 2018
-- Description:  An example of a Moore FSM
-- Author:       J. op den Brouw
-- State:        demo
-- Error:        -
-- Version:      1.3
-- Copyright:    (c)2018, De Haagse Hogeschool

-- This VHDL code describes an example of a Moore
-- finite state machine. There is not one single
-- method to describe a machine but there are two
-- common description: a 3-process and a 2-process
-- machine. Both are presented here below.

-- The 3-process version has three processes:
-- one for generating the Next State Logic (NSL)
-- one for generating the Outpul Logic (OL) and
-- one for generating the state register (REG).

-- The 2-process version has two processes:
-- and one for generating the Next State Logic (NSL)
-- the Outpul Logic (OL),
-- one for generating the state register (REG).

-- The alternative 2-process model incorporates one
-- process for the next state logic and the state register
-- The process for the output logic is identical to the
-- 3-process model.

library ieee;
use ieee.std_logic_1164.all;

entity moore_example is
    port (clk:    in std_logic;
          areset: in std_logic;
          x, y:   in std_logic;
          z:      out std_logic
         );
end moore_example;

-- The 3-process version
architecture fsm_3proc of moore_example is
-- State types
type state_type is (s0, s1, s2);
-- !!!FORCE STATE ENCODING!!!
--attribute ENUM_ENCODING : string;
--attribute ENUM_ENCODING of state_type : type is "00 01 10";
--attribute ENUM_ENCODING of state_type : type is "11 10 00";
--attribute ENUM_ENCODING of state_type : type is "100 010 110";
--attribute ENUM_ENCODING of state_type : type is "001 010 100";
-- Internal current state and next state
signal current_state, next_state : state_type;
begin
    -- Combinational process generates the next state
    combstate: process (current_state, x, y) is
    begin
        case current_state is
            when s0 => if x = '0' or y = '0' then
                           next_state <= s0;
                       else
                           next_state <= s1;
                       end if;
            when s1 => if x = '0' and y = '0' then
                           next_state <= s0;
                       else
                           next_state <= s2;
                       end if;
            when s2 => if x = '0' then
                           next_state <= s0;
                       else
                           next_state <= s2;
                       end if;
            when others => null;
        end case;
    end process;
  
    -- Sequential process generates the
    -- state register
    reg: process (clk, areset) is
    begin
        if areset = '1' then
            current_state <= s0;
        elsif rising_edge(clk) then
            current_state <= next_state;
        end if;
    end process;

    -- Combinational process generates the output
    combout: process (current_state) is
    begin
        case current_state is
            when s0 => z <= '0';
            when s1 => z <= '1';
            when s2 => z <= '1';
            when others => null;
        end case;
    end process;
    -- Or.... much faster...
    --combout: z <= '1' when current_state = s1 or current_state = s2 else '0';
  
end architecture;


-- The 2-process version
architecture fsm_2proc of moore_example is
-- State types
type state_type is (s0, s1, s2);
--attribute ENUM_ENCODING : string;
--attribute ENUM_ENCODING of state_type : type is "00 01 10";
--attribute ENUM_ENCODING of state_type : type is "11 10 00";
--attribute ENUM_ENCODING of state_type : type is "100 010 110";
--attribute ENUM_ENCODING of state_type : type is "001 010 100";
-- Internal current state and next state
signal current_state, next_state : state_type;
begin
    -- Combinational process generates the
    -- next state and the output
    comb: process (current_state, x, y) is
    begin
        case current_state is
            when s0 => z <= '0';
                       if x = '0' or y = '0' then
                           next_state <= s0;
                       else
                           next_state <= s1;
                       end if;
            when s1 => z <= '1';
                       if x = '0' and y = '0' then
                           next_state <= s0;
                       else
                           next_state <= s2;
                       end if;
            when s2 => z <= '1';
                       if x = '0' then
                           next_state <= s0;
                       else
                           next_state <= s2;
                       end if;
            when others => null;
        end case;
    end process;

    -- Sequential process generates the
    -- state register
    reg: process (clk, areset) is
    begin
        if areset = '1' then
            current_state <= s0;
        elsif rising_edge(clk) then
            current_state <= next_state;
        end if;
    end process;

end architecture;

-- The 2-process version (alternative)
architecture fsm_2proc_alt of moore_example is
-- State types
type state_type is (s0, s1, s2);
-- !!!FORCE STATE ENCODING!!!
--attribute ENUM_ENCODING : string;
--attribute ENUM_ENCODING of state_type : type is "00 01 10";
--attribute ENUM_ENCODING of state_type : type is "11 10 00";
--attribute ENUM_ENCODING of state_type : type is "100 010 110";
--attribute ENUM_ENCODING of state_type : type is "001 010 100";
-- Internal (current) state
signal state : state_type;
begin
    -- Sequential process generates the next state and state resgister
    combstate: process (clk, areset) is
    begin
        if areset = '1' then
            state <= s0;
        elsif rising_edge(clk) then
            case state is
                when s0 => if x = '0' or y = '0' then
                               state <= s0;
                           else
                               state <= s1;
                           end if;
                when s1 => if x = '0' and y = '0' then
                               state <= s0;
                           else
                               state <= s2;
                           end if;
                when s2 => if x = '0' then
                               state <= s0;
                           else
                               state <= s2;
                           end if;
                when others => null;
            end case;
    end if;
    end process;
  
    -- Combinational process generates the output
    combout: process (state) is
    begin
        case state is
            when s0 => z <= '0';
            when s1 => z <= '1';
            when s2 => z <= '1';
            when others => null;
        end case;
    end process;
    -- Or.... much faster...
    --combout: z <= '1' when current_state = s1 or current_state = s2 else '0';
  
end architecture;
